require 'socket'
require_relative 'commands/definitions'
require_relative 'commands/hello'
require_relative 'commands/extended_hello'
require_relative 'commands/mail'
require_relative 'commands/recipient'
require_relative 'commands/data'
require_relative 'commands/reset'
require_relative 'commands/quit'

module SMTP
  module_function

  def listen(port, session_klass: SMTP::Session, host: '127.0.0.1')
    @server = TCPServer.new port
    puts "Listening on: #{port}"
    loop do
      Thread.start(session_klass.new(@server.accept, host)) do |session|
        session.run
      end
    end
  end

  class Session
    include Commands::Definitions

    attr_accessor :connection
    attr_accessor :host
    attr_accessor :active
    attr_accessor :domain
    attr_accessor :reverse_path
    attr_accessor :forward_path
    attr_accessor :message

    def initialize(connection, host)
      @connection = connection
      @host = host
      @active = true
      @domain = nil
      setup_session
    end

    def setup_session
      @reverse_path = ''
      @forward_path = ''
      @message = []
    end

    def run
      begin
        @connection.puts reply(220, "#{@host} Simple Mail Transfer")

        while @active 
          @connection.sync = true
          request = @connection.gets
          response = route(process(request))
          @connection.puts "#{response}\r\n"
        end
      rescue => e
        @connection.puts reply(421, e)
        puts e, e.backtrace
      ensure
          @connection.shutdown(:RDWR)
          @connection.close
      end
    end

    def process(request_string)
      if request_string
        request_array = request_string.split(' ')
        {
          command: request_array.shift.to_sym,
          body: request_array.join(' ')
        }
      else
        { command: :NOOP, body: '' }
      end
    end

    def route(request)
      puts "Requesting: #{request[:command]} #{request[:body]}"
      if commands.key?(request[:command])
        response = commands[request[:command]].call(request[:body])
      else
        response = reply(500)
      end
      puts "Responding: #{response}"
      response
    end
  end
end

class MinimalSMTPSession < SMTP::Session
  def hello(body)
    HelloCommand.new(self, body).call
  end
  def extended_hello(body)
    ExtendedHelloCommand.new(self, body).call
  end
  def mail(body)
    MailCommand.new(self, body).call
  end
  def recipient(body)
    RecipientCommand.new(self, body).call
  end
  def data(body)
    DataCommand.new(self, body).call
  end
  def reset(body)
    ResetCommand.new(self, body).call
  end
  def quit(body)
    QuitCommand.new(self, body).call
  end
  def no_operation(body)
    reply(250)
  end
end

# SMTP.listen(5000, session_klass: MinimalSMTPSession)
